﻿namespace WpfApplication2
{
  using System;
  using System.Windows;
  using System.Windows.Media;
  using System.Windows.Media.Imaging;

  public partial class MainWindow : Window
  {
    public MainWindow()
    {
      InitializeComponent();
      this.Loaded += OnLoaded;
    }
    void OnLoaded(object sender, RoutedEventArgs e)
    {
      this.senseManager = PXCMSenseManager.CreateInstance();

      this.senseManager.captureManager.SetRealtime(false);

      // asking for the same image size and frame rate range from both
      // color and depths.
      var frameRate = new PXCMRangeF32() { min = 30, max = 60 };
      var size = new PXCMSizeI32() { width = 640, height = 480 };

      var streamDesc = new PXCMVideoModule.StreamDesc()
      {
        frameRate = frameRate,
        sizeMin = size,
        sizeMax = size
      };

      this.senseManager.EnableStreams(
        new PXCMVideoModule.DataDesc()
        {
          streams = new PXCMVideoModule.StreamDescSet()
          {
            color = streamDesc,
            depth = streamDesc
          }
        }
      ).ThrowOnFail();

      this.senseManager.Init(
        new PXCMSenseManager.Handler()
        {
          onNewSample = this.OnNewSample
        }).ThrowOnFail();

      this.projection = 
        this.senseManager.captureManager.device.CreateProjection();

      this.senseManager.StreamFrames(false);
    }
    pxcmStatus OnNewSample(int mid, PXCMCapture.Sample sample)
    {
      // this is not the UI thread.
      PXCMImage.ImageData depthImage;
      PXCMImage.ImageData colorImage;

      var gotDepth = sample.depth.AcquireAccess(
        PXCMImage.Access.ACCESS_READ, PXCMImage.PixelFormat.PIXEL_FORMAT_DEPTH, out depthImage);

      if (gotDepth.Succeeded())
      {
        var gotColor = sample.color.AcquireAccess(
          PXCMImage.Access.ACCESS_READ_WRITE, PXCMImage.PixelFormat.PIXEL_FORMAT_RGB32, out colorImage);

        if (gotColor.Succeeded())
        {
          this.InitialiseImageDimensions(sample.depth);

          this.QueryUvMapAndConfidenceValue(sample.depth);

          this.ApplyDepthToColor(depthImage, colorImage);

          Dispatcher.InvokeAsync(() =>
            {
              this.InitialiseImage();

              this.writeableBitmap.WritePixels(
                this.imageDimensions,
                colorImage.planes[0],
                this.imageDimensions.Width * this.imageDimensions.Height * 4,
                this.imageDimensions.Width * 4);

              sample.color.ReleaseAccess(colorImage);
              sample.depth.ReleaseAccess(depthImage);
            }
          );
        }
        else
        {
          sample.depth.ReleaseAccess(depthImage);
        }
      }
      return (pxcmStatus.PXCM_STATUS_NO_ERROR);
    }
    void QueryUvMapAndConfidenceValue(PXCMImage depthImage)
    {
      if (this.uvMap == null)
      {
        this.uvMap = new PXCMPointF32[this.imageDimensions.Width * this.imageDimensions.Height];
        this.depthLowConfidenceValue = this.senseManager.captureManager.device.QueryDepthLowConfidenceValue();
      }
      this.projection.QueryUVMap(depthImage, this.uvMap).ThrowOnFail();
    }
    void InitialiseImageDimensions(PXCMImage image)
    {
      if (!this.imageDimensions.HasArea)
      {
        this.imageDimensions.Width = image.info.width;
        this.imageDimensions.Height = image.info.height;
      }
    }
    void InitialiseImage()
    {
      if (this.writeableBitmap == null)
      {
        this.writeableBitmap = new WriteableBitmap(
          this.imageDimensions.Width,
          this.imageDimensions.Height,
          96,
          96,
          PixelFormats.Bgra32,
          null);

        this.displayImage.Source = this.writeableBitmap;     
      }
    }
    void ApplyDepthToColor(
      PXCMImage.ImageData depthData,
      PXCMImage.ImageData colorData)
    {
      unsafe
      {
        UInt16* depthPtr = (UInt16*)depthData.planes[0].ToPointer();
        UInt32* colorPtr = (UInt32*)colorData.planes[0].ToPointer();

        var width = this.imageDimensions.Width;
        var height = this.imageDimensions.Height;
        var length = width * height;

        for (int i = 0; i < length; i++)
        {
          colorPtr[i] &= 0x00FFFFFF;
        }

        for (int y = 0; y < height; y++)
        {
          for (int x = 0; x < width; x++)
          {
            int pixelIndex = (y * width) + x;

            if (depthPtr[pixelIndex] != this.depthLowConfidenceValue)
            {
              // tbd - not sure I understand this yet but this map takes a
              // u,v in the depth image and gives a 0..1.0,0..1.0 value
              // to index into the width,height of the colour image.
              // sometimes it comes back as > 1.0 though and sometimes
              // it seems to come back as -1.0?
              var mapped = this.uvMap[pixelIndex];
              mapped.x = Math.Min(1.0f, mapped.x);
              mapped.y = Math.Min(1.0f, mapped.y);

              if ((mapped.x >= 0) && (mapped.y >= 0))
              {
                var cx = (int)(mapped.x * width);
                var cy = (int)(mapped.y * height);
                colorPtr[(cy * width) + cx] |= 0xCC000000;
              }
            }
          }
        }
      }
    }
    UInt16 depthLowConfidenceValue;
    PXCMPointF32[] uvMap;
    Int32Rect imageDimensions;
    WriteableBitmap writeableBitmap;
    PXCMSenseManager senseManager;
    PXCMProjection projection;
  }
}